#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <zlib.h>

#include <rxi/log.h>
#include <rxi/vec.h>

#include <cobble/nbt.h>
#include <cobble/util.h>

/* size of the blocks of memory we allocate for reading in the compressed data */
#define CHUNK_SIZE 1024

static bool
gzip_decompress_data(uint8_t *data, size_t data_len, uint8_t **dest, size_t *dest_len)
{
	/* initialize the variables, otherwise bad things can happen */
	*dest = NULL;
	*dest_len = 0;

	z_stream stream = {
        .zalloc   = Z_NULL,
        .zfree    = Z_NULL,
        .opaque   = Z_NULL,
        .next_in  = data,
        .avail_in = data_len
    };

	if(inflateInit2(&stream, 15 + 32) != Z_OK) {
		goto decompression_error;
	}

	do {
		if (*dest == NULL) {
			*dest = SAFE_MALLOC(CHUNK_SIZE);
		} else {
			SAFE_REALLOC(*dest, *dest_len + CHUNK_SIZE);
		}
		
		stream.avail_out = CHUNK_SIZE;
		stream.next_out = *dest + *dest_len;

		switch (inflate(&stream, Z_NO_FLUSH)) {
		case Z_MEM_ERROR:
		case Z_DATA_ERROR:
		case Z_NEED_DICT:
			goto decompression_error;
		default:
			*dest_len += (CHUNK_SIZE - stream.avail_out);
			break;
		}

	} while (stream.avail_out == 0);

	(void)inflateEnd(&stream);
	return true;
	
decompression_error:
	log_error("Could not decompress NBT data.");
	return false;
}

/* recursive function used to read in an NBTNode, and all child nodes */
static size_t 
nbt_parse_node(struct nbt_node *node, uint8_t *data, size_t data_len)
{
	/* keeps track of the number of bytes we have read */
	size_t size;

	size = 0;
	node->name_len = 0;
	node->name = NULL;

	node->type = (enum nbt_type) data[0];
	size++;

	/* if this is an end tag, there is no need to read in the name */
	if (node->type == TAG_END) {
		return size;
	}

	node->name_len = ((int16_t*)(data + size))[0];
	swap_endian(&node->name_len, 2);
	size += 2;

	if (node->name_len == 0) {
		node->name = NULL;
	} else {
		node->name = SAFE_MALLOC(node->name_len + 1);
		memcpy(node->name, data + size, node->name_len);
		node->name[node->name_len] = '\0';
		size += node->name_len;
	}

	switch (node->type) {
	case TAG_BYTE:
		node->payload.tag_byte = ((int8_t*)(data + size))[0];
		size += 1;
		break;
	case TAG_SHORT:
		node->payload.tag_short = ((int16_t*)(data + size))[0];
		swap_endian(&node->payload.tag_short, 2);
		size += 2;
		break;
	case TAG_INT:
		node->payload.tag_int = ((int32_t*)(data + size))[0];
		swap_endian(&node->payload.tag_int, 4);
		size += 4;
		break;
	case TAG_LONG:
		node->payload.tag_long = ((int64_t*)(data + size))[0];
		swap_endian(&node->payload.tag_long, 8);
		size += 8;
		break;
	case TAG_FLOAT:
		node->payload.tag_float = ((float*)(data + size))[0];
		swap_endian(&node->payload.tag_float, 4);
		size += 4;
		break;
	case TAG_DOUBLE:
		node->payload.tag_long = ((double*)(data + size))[0];
		swap_endian(&node->payload.tag_double, 8);
		size += 8;
		break;
	case TAG_STRING:
		{
			size_t len;

			len = ((int16_t*)(data + size))[0];
			swap_endian(&len, 2);
			size += 2;

			if (len == 0){
				node->payload.tag_string = NULL;
			} else {
				node->payload.tag_string = SAFE_MALLOC(len + 1);
				memcpy(node->payload.tag_string, data + size, len);
				node->payload.tag_string[len] = '\0';
				size += len;
			}
		}
		break;
	case TAG_COMPOUND: 
		vec_init(&node->payload.tag_compound);
	
		while (1) {
			struct nbt_node *child_node;

			child_node = SAFE_MALLOC(sizeof(struct nbt_node));
			size += nbt_parse_node(child_node, data + size, data_len - size);
			vec_push(&node->payload.tag_compound, child_node);

			if (child_node->type == TAG_END){
				break;
			}
		}
		break;
	default:
		/* TODO remove this */
		log_error("This NBT type has not been implimented");
	}

	return size;
}

bool
nbt_parse_file(struct nbt_node *node, FILE *fp)
{
	void *data;
	size_t file_len;
	
	fseek(fp, 0, SEEK_END); /* jump to the end of the file */
	file_len = ftell(fp);
	rewind(fp); /* jump back to the beginning of the file */
	data = SAFE_MALLOC(file_len);
	fread(data, file_len, 1, fp);

	nbt_parse_bytes(node, data, file_len);
	free(data);
	return true;
}

bool
nbt_parse_bytes(struct nbt_node *node, uint8_t *data, size_t data_len)
{
	uint8_t *idata;
	size_t idata_len;

	if (!gzip_decompress_data(data, data_len, &idata, &idata_len)) {
		return false;
	}

	nbt_parse_node(node, idata, idata_len);

	free(idata);
	return true;		
}

struct nbt_node* 
nbt_node_get(struct nbt_node node, char* name)
{
	struct nbt_node *val;
	int i;

	if (node.type != TAG_COMPOUND) return NULL;

	vec_foreach(&node.payload.tag_compound, val, i){
		if (strcmp(node.name, val->name) == 0) {
			return val;
		}
	}
	return NULL;
}

void
nbt_node_destroy(struct nbt_node* node)
{
	free(node->name);
	
	switch (node->type) {
	case TAG_STRING:
		free(node->payload.tag_string);
		break;
	case TAG_COMPOUND:
		{
			int i;
			struct nbt_node *val;
			vec_foreach(&node->payload.tag_compound, val, i) {
				nbt_node_destroy(val);
			}
		}
		break;
	}
	free(node);
}
