/* see LICENSE file for copyright and license details. */
#include <stdbool.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>

#include <rxi/log.h>

#include <cobble/connection.h>
#include <cobble/packet.h>
#include <cobble/packet_impl.h>
#include <cobble/player.h>
#include <cobble/protocol.h>
#include <cobble/server.h>
#include <cobble/world.h>
#include <cobble/util.h>

#include <../config.h>

/* TODO: make protocol functions actually error check */

/* HANDSHAKE_SERVER_HANDSHAKE =============================================== */
bool
packet_handshake_read(struct connection *conn, struct packet *packet, uint8_t *buf, size_t len)
{
	struct packet_handshake *p;
	size_t size;

	size = 0;
	p = SAFE_MALLOC(sizeof(struct packet_handshake));

	size += read_varint(&p->protocol, buf + size, len - size);
	size += read_string(&p->server_address, buf + size, len - size);
	size += read_ushort(&p->server_port, buf + size, len - size);
	size += read_varint(&p->next_state, buf + size, len - size);

	packet->data = (void*) p;
	return true;
}

void
packet_handshake_destroy(struct packet *p)
{
	struct packet_handshake *packet;

	packet = (struct packet_handshake*) p->data;
	free(packet->server_address);
}

void
packet_handshake_handle(struct connection *conn, struct packet *p)
{
	struct packet_handshake *packet;

	packet = (struct packet_handshake*) p->data;

	conn->protocol = packet->protocol;	

	/* TODO block certain protocols */
	if (packet->next_state == 1) {
		conn->protocol_state = STATUS;
	} else if (packet->next_state == 2) {
		conn->protocol_state = LOGIN;
	}
}

/* PLAY_CLIENT_JOIN_GAME ---------------------------------------------------= */
bool
packet_join_game_write(struct connection *conn, struct packet *p, uint8_t *data, size_t data_len)
{
	struct packet_join_game *packet;
	size_t size;

	packet = (struct packet_join_game*) p->data;
	size = 0;
	
	size += write_int(packet->entity_id, data + size);
	(data + size++)[0] = packet->gamemode;
	size += write_int(packet->dimension, data + size);
	(data + size++)[0] = (uint8_t) packet->difficulty;
	(data + size++)[0] = packet->max_players;
	size += write_string(packet->level_type, data + size);
	write_bool(packet->reduced_debug_info, data + size);

	return true;
}

size_t
packet_join_game_size(struct packet *packet)
{
	/* 12 is the size of all the fields but the level_type string */
	return 12 + get_string_len(((struct packet_join_game*) packet->data)->level_type);
}

void
packet_join_game_destroy(struct packet *p)
{
	struct packet_join_game *packet;

	packet = (struct packet_join_game*) p->data;
	free(packet->level_type);
}

/* STATUS_CLIENT_RESPONSE =================================================== */
bool
packet_response_write(struct connection *conn, struct packet *packet, uint8_t *buf, size_t len)
{
	struct packet_response *p;

	p = (struct packet_response*) packet->data;
	write_string(p->response, buf);

	return true;
}

size_t
packet_response_size(struct packet *packet)
{
	return get_string_len(((struct packet_response*) packet->data)->response);
}

void 
packet_response_destroy(struct packet *packet)
{
	struct packet_response *p;

	p = (struct packet_response*) packet->data;
	free(p->response);
}


/* STATUS_CLIENT_PONG ======================================================= */
bool
packet_pong_write(struct connection *conn, struct packet *packet, uint8_t *buf, size_t len)
{
	struct packet_pong *p;

	p = (struct packet_pong*) packet->data;			
	write_long(p->payload, buf);

	return true;
}

size_t
packet_pong_size(struct packet *packet)
{
	return 8;
}

/* STATUS_SERVER_REQUEST ==================================================== */
#define RESPONSE_BUFFER_SIZE 1024

bool
packet_request_read(struct connection *conn, struct packet* packet, uint8_t *data, size_t data_len)
{
	return true;
}

void
packet_request_handle(struct connection *conn, struct packet *p) 
{
	struct packet_response *packet_data;
	struct packet *new_packet;

	packet_data = SAFE_MALLOC(sizeof(struct packet_response));

	packet_data->response = SAFE_MALLOC(sizeof(char) * RESPONSE_BUFFER_SIZE);
	packet_data->response[RESPONSE_BUFFER_SIZE - 1] = 0;
	/* TODO actually get all of this */
	snprintf(packet_data->response, 
			RESPONSE_BUFFER_SIZE, 
			"{\"version\":{\"name\":\"1.14.4\",\"protocol\":498},\"players\":{\"max\":%d,\"online\":0},\"description\":{\"text\":\"%s\"}}", 
			COBBLE_MAX_PLAYERS,
			COBBLE_SERVER_MOTD);

	new_packet = packet_new(STATUS_CLIENT_RESPONSE, packet_data);
	if (!packet_write(conn, new_packet)) {
		log_error("Failed to write STATUS_CLIENT_RESPONSE packet");
	}
	packet_destroy(new_packet);
}

/* STATUS_SERVER_PING ======================================================= */
bool
packet_ping_read(struct connection *conn, struct packet *packet, uint8_t *buf, size_t len)
{
	struct packet_ping *p;

	p = SAFE_MALLOC(sizeof(struct packet_ping));
	read_long(&p->payload, buf, len);
	packet->data = (void*) p;

	return true;
}

void
packet_ping_handle(struct connection *conn, struct packet *p)
{
	struct packet_ping *ping_data;
	struct packet_pong *pong_data;
	struct packet *new_packet;

	ping_data = (struct packet_ping*) p->data;
	pong_data = SAFE_MALLOC(sizeof(struct packet_pong));

	pong_data->payload = ping_data->payload;

	new_packet = packet_new(STATUS_CLIENT_PONG, pong_data);

	if (!packet_write(conn, new_packet)) {
		log_error("Failed to write STATUS_CLIENT_RESPONSE packet");
	}
	packet_destroy(new_packet);
}

/* LOGIN_CLIENT_LOGIN_SUCCESS =============================================== */
bool
packet_login_success_write(struct connection *conn, struct packet *p, uint8_t *buf, size_t data_len)
{
	struct packet_login_success *packet;
	size_t size;

	packet = (struct packet_login_success*) p->data;
	size = 0;

	size += write_string(packet->uuid, buf);
	size += write_string(packet->username, buf + size);

	return true;
}

size_t
packet_login_success_size(struct packet *p)
{
	struct packet_login_success *packet;
	size_t size;

	packet = (struct packet_login_success*) p->data;
	size = 0;
	size += get_string_len(packet->uuid);
	size += get_string_len(packet->username);

	return size;
}

void
packet_login_success_destroy(struct packet *p)
{
	struct packet_login_success *packet;

	packet = (struct packet_login_success*) p->data;

	free(packet->uuid);
	free(packet->username);
}

/* LOGIN_SERVER_LOGIN_START ================================================= */
bool
packet_login_start_read(struct connection *conn, struct packet *p, uint8_t *buf, size_t len)
{
	struct packet_login_start *packet;
	
	packet = SAFE_MALLOC(sizeof(struct packet_login_start));
	read_string(&packet->username, buf, len);

	p->data = packet;

	return true;
}

void
packet_login_start_destroy(struct packet *p)
{
	struct packet_login_start *packet;

	packet = (struct packet_login_start*) p->data;
	free(packet->username);
}

void
packet_login_start_handle(struct connection *conn, struct packet *p)
{
	struct packet_login_start *login_start_data;
	login_start_data = (struct packet_login_start*) p->data;

	if (COBBLE_ONLINE_MODE) {
		log_error("Online mode login isn't yet implimented!");
	} else {
		struct packet_login_success *login_success_data;
		struct packet *new_packet;
		struct uuid *uuid;

		login_success_data = SAFE_MALLOC(sizeof(struct packet_login_success));
		uuid = SAFE_MALLOC(sizeof(struct uuid));

		/* create a uuid by md5 hashing the players username, then encode it into a string */
		generate_uuid(login_start_data->username, uuid);	
		login_success_data->uuid = SAFE_MALLOC(38);
		snprintf(login_success_data->uuid     , 10, "%08X-", ((uint32_t*) uuid)[0]);
		snprintf(login_success_data->uuid + 9 , 6 , "%04X-", ((uint16_t*) uuid)[2]);
		snprintf(login_success_data->uuid + 14, 6 , "%04X-", ((uint16_t*) uuid)[3]);
		snprintf(login_success_data->uuid + 19, 6 , "%04X-", ((uint16_t*) uuid)[4]);
		snprintf(login_success_data->uuid + 24, 9 , "%08X" , ((uint32_t*) (uuid + 4))[2]);
		snprintf(login_success_data->uuid + 32, 5 , "%04X" , ((uint16_t*) uuid)[7]);

		login_success_data->username = SAFE_MALLOC(strlen(login_start_data->username));
		strcpy(login_success_data->username, login_start_data->username);

		new_packet = packet_new(LOGIN_CLIENT_LOGIN_SUCCESS, login_success_data);

		if (!packet_write(conn, new_packet)) {
			log_error("Failed to write LOGIN_CLIENT_LOGIN_SUCCESS packet");
		}
		free(uuid);
		packet_destroy(new_packet);
	}
}
