/* see LICENSE file for copyright and license details. */
#include <byteswap.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <openssl/md5.h>

#include <cobble/protocol.h>
#include <cobble/util.h>

#define MAX_STRING_LENGTH 32767  

size_t
read_long(int64_t *output, uint8_t *buf, size_t buf_len) 
{
	*output = 0;
	if (buf_len < 8) return 0;

	for (int i = 0;i < 8;i++) {
		*output |= buf[i] >> (i * 8);
	}
	return 8;
}

size_t
read_int(int32_t *output, uint8_t *buf, size_t buf_len) 
{
	*output = 0;	
	if (buf_len < 4) return 0;

	memcpy(output, buf, 4);
	*output = __bswap_32(*output);

	return 4;
}

size_t
read_ushort(uint16_t *output, uint8_t *buf, size_t buf_len) 
{
	*output = 0;	
	if (buf_len < 2) return 0;

	memcpy(output, buf, 2);
	*output = __bswap_16(*output);

	return 2;
}

size_t
read_varint(int32_t *output, uint8_t *buf, size_t buf_len) 
{
	*output = 0;
    int bytes_read = 0;
    signed char read;
    do {
        if (bytes_read >= buf_len) return 0;

        read = buf[bytes_read];
        *output |= (read & 127) << (bytes_read++ * 7);

        if (bytes_read > 5) return bytes_read;
    } while ((read & 128) == 128);

	return bytes_read;
}

size_t
read_varlong(int64_t *output, uint8_t *buf, size_t buf_len)
{
	*output = 0;
    int bytes_read = 0;
    signed char read;
    do {
        if (bytes_read >= buf_len) return 0;

        read = buf[bytes_read];
        *output |= (read & 127) << (bytes_read++ * 7);

        if (bytes_read > 10) return bytes_read;
    } while ((read & 128) == 128);

	return bytes_read;
}

size_t
read_string(char** output, uint8_t *buf, size_t buf_len) 
{
	int32_t string_length, size;

	if (buf_len == 0) goto fail;
	size = read_varint(&string_length, buf, buf_len);
	if (size == 0) goto fail;
	if (string_length > MAX_STRING_LENGTH) goto fail;

	buf += size;
	buf_len -= size;
	if (buf_len < string_length) goto fail;

	*output = SAFE_MALLOC(sizeof(char) * (string_length + 1));
	memcpy(*output, buf, string_length);
	(*output)[string_length] = 0;
	
	return string_length + size;

	fail:
	// create an empty string
	*output = SAFE_MALLOC(sizeof(char));
	*output[0] = 0;
	return 0;
}

size_t
read_bool(bool *output, uint8_t *buf, size_t buf_len)
{
	if (buf_len < 1) return -1;
	*output = (buf[0] == 0x01);

	return 1;
}

size_t write_long(int64_t input, uint8_t *buf)
{
	for (int i = 0;i < 8;i++) {
		buf[i] = input >> (8 * i) & 0xFF;
	}
	return 8;
}

size_t write_int(int32_t input, uint8_t *buf)
{
	for (int i = 0;i < 4;i++) {
		buf[i] = input >> (8 * i) & 0xFF;
	}
	return 4;
}

size_t
write_ushort(uint16_t input, uint8_t *buf)
{
	for (int i = 0;i < 2;i++) {
		buf[i] = input >> (8 * i) & 0xFF;
	}
	return 2;
}

size_t
write_varint(int32_t input, uint8_t *buf)
{
	int i = 0;
	while ((input & -128) != 0) {
	    buf[i++] = (input & 127) | 128;
	    input >>= 7;
	}
	buf[i++] = input;
	return i;
}

size_t
write_varlong(int64_t input, uint8_t *buf)
{
	int i = 0;
	while ((input & -128) != 0) {
	    buf[i++] = (input & 127) | 128;
	    input >>= 7;
	}
	buf[i++] = input;
	return i;
}

size_t
write_string(char *input, uint8_t *buf)
{
	size_t str_len;
	uint8_t int_len;

	str_len = strlen(input);
	int_len = write_varint((int32_t) str_len, buf);
	memcpy(buf + int_len, input, str_len);

	return str_len + int_len;
}

size_t
write_bool(bool input, uint8_t *buf)
{
	if (input) {
		buf[0] = 0x01;
	} else {
		buf[1] = 0x00;
	}
	return 1;
}

size_t
get_varint_len(int32_t input) 
{
    for (int32_t x = 1; x < 5; x++) {
        if ((input & -1 << x * 7) == 0) return x;
    }
    return 5;
}

size_t
get_varlong_len(int64_t input) 
{
    for (int32_t x = 1; x < 10; x++) {
        if ((input & -1 << x * 7) == 0) return x;
    }
    return 10;
}

size_t
get_string_len(char* input) 
{
	size_t size;

	size = strlen(input);
	return size + get_varint_len(size); 
}

bool
generate_uuid(char *username, struct uuid *uuid)
{
	MD5_CTX context;
	MD5_Init(&context);	
	MD5_Update(&context, username, strlen(username));
	MD5_Final((void*) uuid, &context);

	return true;
}
